import torch
import torch.nn as nn
import torch.nn.functional as F
from layers import GraphAttentionLayer, SpGraphAttentionLayer, globalAttention
import numpy as np

class GAT(nn.Module):
    def __init__(self, nfeat, nhid, nclass, dropout, alpha, nheads):
        """Dense version of GAT."""
        super(GAT, self).__init__()
        self.dropout = dropout

        self.attentions = [GraphAttentionLayer(nfeat, nhid, dropout=dropout, alpha=alpha, concat=True) for _ in range(nheads)]
        for i, attention in enumerate(self.attentions): # first layer consisting of nheads.
            self.add_module('attention_{}'.format(i), attention)

        self.out_att = GraphAttentionLayer(nhid * nheads, nclass, dropout=dropout, alpha=alpha, concat=False) # second layer

    def forward(self, x, adj):
        x = F.dropout(x, self.dropout, training=self.training)
        x = torch.cat([att(x, adj) for att in self.attentions], dim=1)
        x = F.dropout(x, self.dropout, training=self.training)
        x = F.elu(self.out_att(x, adj))
        return F.log_softmax(x, dim=1)


class SpGAT(nn.Module):
    def __init__(self, nfeat, nhid, outDim, nclass, dropout, alpha, JKNet):
        """Sparse version of GAT."""
        super(SpGAT, self).__init__()
        self.dropout = dropout
        self.nclass = nclass
        self.nhid = nhid
        self.outDim = outDim
        self.JKNet = JKNet
        self.attentions = [SpGraphAttentionLayer(in_features=nfeat,
                                                 out_features=nhid[0],
                                                 nclass=nclass,
                                                 dropout=dropout, 
                                                 alpha=alpha, 
                                                 concat=True) for _ in range(nclass)]
        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)


        self.attentions_layer1 = SpGraphAttentionLayer(in_features=nhid[0],
                                             out_features=nhid[1],
                                             nclass=nclass,
                                             dropout=dropout,
                                             alpha=alpha,
                                             concat=True, repeat=nclass)
        self.add_module('attention_layer{}'.format(2), self.attentions_layer1)

        self.attentions_layer2 = SpGraphAttentionLayer(in_features=nhid[1],
                                              out_features=nhid[2],
                                              nclass=nclass,
                                              dropout= dropout,
                                              alpha=alpha,
                                              concat=True, repeat=nclass)
        self.add_module('attention_layer{}'.format(3), self.attentions_layer2)

        if self.JKNet:
            self.attentions_global = globalAttention(in_features=np.sum(nhid), out_features=outDim, nclass=nclass,
                                                     isactive=False)
        else:
            self.attentions_global = globalAttention(in_features=nhid[-1], out_features=outDim, nclass=nclass,isactive=False)

        self.add_module('attention_global', self.attentions_global)
        # self.add_module('attention_layer2', self.out_att)
        # self.out_att2 = SpGraphAttentionLayer(in_features=nhid[1],
        #                                      out_features=nhid[2],
        #                                      nclass=nclass,
        #                                      dropout=dropout,
        #                                      alpha=alpha,
        #                                      concat=False, repeat=nclass)

    def forward(self, x, adj):
        # x = F.dropout(x, self.dropout, training=self.training) ## None

        middleResults = []

        # first layer
        x = torch.cat([att(x, adj) for att in self.attentions], dim=1) #
        x = F.dropout(x, self.dropout, training=self.training) # N x d1

        # skip connection
        E_layer1 = self.reconstructForm(x.clone(), self.nhid[0])
        middleResults.append(E_layer1)

        # second layer
        x = self.attentions_layer1(x, adj)
        x = F.dropout(x, self.dropout, training=self.training)

        E_layer2 = self.reconstructForm(x.clone(), self.nhid[1]) # nclass x N x d
        middleResults.append(E_layer2)

        # third layer
        x = self.attentions_layer2(x, adj)
        x = F.dropout(x, self.dropout, training=self.training)

        E_layer3 = self.reconstructForm(x, self.nhid[2]) # nclass x N x dim

        if self.JKNet:
            middleResults.append(E_layer3)
            # contact JK-Net
            results = torch.cat(middleResults, dim=2) # nclass x N x (sum(nhid))
        else:
            results = E_layer3

        middleResults = torch.cat([E_layer1[i, :, :] for i in range(self.nclass)], dim=1)
        # full connection to learn weight
        global_embedding = self.attentions_global(results)
        global_embedding = F.log_softmax(global_embedding,dim = 2)

        return global_embedding, middleResults


    def reconstructForm(self, input, perdim):
        """reconstructure input into a tensor"""
        # E = torch.Tensor([]).cuda()
        E = torch.Tensor([]).cuda()
        for xi in range(self.nclass):
            E_part = input[:, xi * perdim: (xi + 1) * perdim]
            E = torch.cat((E, E_part.unsqueeze(dim=0)), dim=0)
        return E